"use strict";
// tslint:disable:member-ordering
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var d3_array_1 = require("d3-array");
var d3_collection_1 = require("d3-collection");
var d3_scale_1 = require("d3-scale");
var reselect_1 = require("reselect");
var colors_1 = require("./colors");
var FlowMapLayer_1 = __importDefault(require("./FlowMapLayer"));
var types_1 = require("./types");
var CIRCLE_OUTLINE_THICKNESS = 1;
var getDiffMode = function (props) { return props.diffMode; };
var getColorsProp = function (props) { return props.colors; };
var getAnimate = function (props) { return props.animate; };
var getLocationFeatures = function (props) {
    return types_1.isFeatureCollection(props.locations) ? props.locations.features : props.locations;
};
var getFlows = function (props) { return props.flows; };
var getHighlightedFlow = function (props) { return props.highlightedFlow; };
var getHighlightedLocationId = function (props) { return props.highlightedLocationId; };
var getSelectedLocationIds = function (props) { return props.selectedLocationIds; };
var getShowOnlyTopFlows = function (props) { return props.showOnlyTopFlows; };
var getMaxLocationCircleSize = function (props) {
    return props.maxLocationCircleSize != null ? props.maxLocationCircleSize : FlowMapLayer_1.default.defaultProps.maxLocationCircleSize;
};
var Selectors = /** @class */ (function () {
    function Selectors(inputAccessors) {
        var _this = this;
        this.inputAccessors = inputAccessors;
        this.getColors = reselect_1.createSelector([getColorsProp, getDiffMode], function (colors, diffMode) {
            if (diffMode) {
                return colors_1.getDiffColorsRGBA(colors);
            }
            return colors_1.getColorsRGBA(colors);
        });
        this.getLocationByIdGetter = reselect_1.createSelector([getLocationFeatures], function (locations) {
            var locationsById = d3_collection_1.nest()
                .key(_this.inputAccessors.getLocationId)
                .rollup(function (_a) {
                var d = _a[0];
                return d;
            })
                .object(locations);
            return function (id) {
                var location = locationsById[id];
                if (!location) {
                    console.warn("No location found for id '" + id + "'");
                }
                return location;
            };
        });
        this.getFilteredFlows = reselect_1.createSelector([getFlows, getSelectedLocationIds], function (flows, selectedLocationIds) {
            var _a = _this.inputAccessors, getFlowOriginId = _a.getFlowOriginId, getFlowDestId = _a.getFlowDestId;
            if (selectedLocationIds) {
                return flows.filter(function (flow) {
                    return selectedLocationIds.indexOf(getFlowOriginId(flow)) >= 0 ||
                        selectedLocationIds.indexOf(getFlowDestId(flow)) >= 0;
                });
            }
            return flows;
        });
        this.getNonSelfFlows = reselect_1.createSelector([this.getFilteredFlows], function (flows) {
            var _a = _this.inputAccessors, getFlowOriginId = _a.getFlowOriginId, getFlowDestId = _a.getFlowDestId;
            return flows.filter(function (flow) { return getFlowOriginId(flow) !== getFlowDestId(flow); });
        });
        this.getSortedNonSelfFlows = reselect_1.createSelector([this.getNonSelfFlows], function (flows) {
            var comparator = function (f1, f2) {
                return Math.abs(_this.inputAccessors.getFlowMagnitude(f1)) - Math.abs(_this.inputAccessors.getFlowMagnitude(f2));
            };
            return flows.slice().sort(comparator);
        });
        this.getTopFlows = reselect_1.createSelector([this.getSortedNonSelfFlows, getShowOnlyTopFlows], function (flows, showOnlyTopFlows) {
            if (showOnlyTopFlows != null && showOnlyTopFlows > 0 && flows.length > showOnlyTopFlows) {
                return flows.slice(flows.length - showOnlyTopFlows, flows.length);
            }
            return flows;
        });
        this.getHighlightedFlows = reselect_1.createSelector([this.getSortedNonSelfFlows, getHighlightedFlow, getHighlightedLocationId], function (flows, highlightedFlow, highlightedLocationId) {
            var _a = _this.inputAccessors, getFlowOriginId = _a.getFlowOriginId, getFlowDestId = _a.getFlowDestId;
            if (highlightedFlow) {
                return [highlightedFlow];
            }
            if (highlightedLocationId) {
                return flows.filter(function (flow) { return getFlowOriginId(flow) === highlightedLocationId || getFlowDestId(flow) === highlightedLocationId; });
            }
            return undefined;
        });
        this.getFlowMagnitudeExtent = reselect_1.createSelector([this.getNonSelfFlows, function (props) { return props.flowMagnitudeExtent; }], function (flows, flowMagnitudeExtent) {
            if (flowMagnitudeExtent != null)
                return flowMagnitudeExtent;
            return d3_array_1.extent(flows, function (f) { return _this.inputAccessors.getFlowMagnitude(f); });
        });
        this.getFlowThicknessScale = reselect_1.createSelector([this.getFlowMagnitudeExtent], function (_a) {
            var minMagnitude = _a[0], maxMagnitude = _a[1];
            var scale = d3_scale_1.scaleLinear()
                .range([0.05, 0.5])
                .domain([0, Math.max(Math.abs(minMagnitude || 0), Math.abs(maxMagnitude || 0))]);
            return function (magnitude) { return scale(Math.abs(magnitude)); };
        });
        this.getFlowColorScale = reselect_1.createSelector([this.getColors, this.getFlowMagnitudeExtent, getAnimate], function (colors, _a, animate) {
            var minMagnitude = _a[0], maxMagnitude = _a[1];
            if (colors_1.isDiffColorsRGBA(colors)) {
                var posScale_1 = colors_1.createFlowColorScale([0, maxMagnitude || 0], colors.positive.flows.scheme, animate);
                var negScale_1 = colors_1.createFlowColorScale([0, minMagnitude || 0], colors.negative.flows.scheme, animate);
                return function (magnitude) { return (magnitude >= 0 ? posScale_1(magnitude) : negScale_1(magnitude)); };
            }
            var scale = colors_1.createFlowColorScale([0, maxMagnitude || 0], colors.flows.scheme, animate);
            return function (magnitude) { return scale(magnitude); };
        });
        this.getLocationTotals = reselect_1.createSelector([getLocationFeatures, this.getFilteredFlows], function (locations, flows) {
            var _a = _this.inputAccessors, getFlowOriginId = _a.getFlowOriginId, getFlowDestId = _a.getFlowDestId, getFlowMagnitude = _a.getFlowMagnitude;
            return flows.reduce(function (acc, curr) {
                var originId = getFlowOriginId(curr);
                var destId = getFlowDestId(curr);
                var magnitude = getFlowMagnitude(curr);
                if (originId === destId) {
                    acc.within[originId] = (acc.within[originId] || 0) + magnitude;
                }
                else {
                    acc.outgoing[originId] = (acc.outgoing[originId] || 0) + magnitude;
                    acc.incoming[destId] = (acc.incoming[destId] || 0) + magnitude;
                }
                return acc;
            }, { incoming: {}, outgoing: {}, within: {} });
        });
        this.getHighlightedLocationCircles = reselect_1.createSelector([this.getLocationByIdGetter, getHighlightedLocationId], function (getLocationById, highlightedLocationId) {
            if (highlightedLocationId) {
                var location_1 = getLocationById(highlightedLocationId);
                if (!location_1) {
                    return undefined;
                }
                return [
                    { location: location_1, type: "outline" /* OUTLINE */ },
                    { location: location_1, type: "outer" /* OUTER */ },
                    { location: location_1, type: "inner" /* INNER */ },
                ];
            }
            return undefined;
        });
        this.getLocationTotalInGetter = function (props) {
            var _a = _this.inputAccessors, getLocationTotalIn = _a.getLocationTotalIn, getLocationId = _a.getLocationId;
            if (getLocationTotalIn) {
                return getLocationTotalIn;
            }
            var incoming = _this.getLocationTotals(props).incoming;
            return function (location) { return incoming[getLocationId(location)] || 0; };
        };
        this.getLocationTotalOutGetter = function (props) {
            var _a = _this.inputAccessors, getLocationTotalOut = _a.getLocationTotalOut, getLocationId = _a.getLocationId;
            if (getLocationTotalOut) {
                return getLocationTotalOut;
            }
            var outgoing = _this.getLocationTotals(props).outgoing;
            return function (location) { return outgoing[getLocationId(location)] || 0; };
        };
        this.getLocationTotalWithinGetter = function (props) {
            var _a = _this.inputAccessors, getLocationTotalWithin = _a.getLocationTotalWithin, getLocationId = _a.getLocationId;
            if (getLocationTotalWithin) {
                return getLocationTotalWithin;
            }
            var within = _this.getLocationTotals(props).within;
            return function (location) { return within[getLocationId(location)] || 0; };
        };
        this.getLocationMaxAbsTotalGetter = reselect_1.createSelector([
            getLocationFeatures,
            this.getLocationTotalInGetter,
            this.getLocationTotalOutGetter,
            this.getLocationTotalWithinGetter,
        ], function (locations, getLocationTotalIn, getLocationTotalOut, getLocationTotalWithin) {
            return function (location) {
                return Math.max(Math.abs(getLocationTotalIn(location) + getLocationTotalWithin(location)), Math.abs(getLocationTotalOut(location) + getLocationTotalWithin(location)));
            };
        });
        this.getMaxLocationMaxAbsTotal = reselect_1.createSelector([getLocationFeatures, this.getLocationMaxAbsTotalGetter, function (props) { return props.locationTotalsExtent; }], function (locations, getLocationMaxAbsTotal, locationTotalsExtent) {
            if (locationTotalsExtent != null) {
                return d3_array_1.max(locationTotalsExtent, Math.abs) || 0;
            }
            return d3_array_1.max(locations, getLocationMaxAbsTotal) || 0;
        });
        this.getLocationCircles = reselect_1.createSelector([getLocationFeatures, this.getLocationMaxAbsTotalGetter], function (locations, getLocationMaxAbsTotalGetter) {
            var circles = [];
            var sorted = locations
                .slice()
                .sort(function (a, b) { return d3_array_1.ascending(getLocationMaxAbsTotalGetter(a), getLocationMaxAbsTotalGetter(b)); });
            for (var _i = 0, sorted_1 = sorted; _i < sorted_1.length; _i++) {
                var location_2 = sorted_1[_i];
                circles.push({
                    location: location_2,
                    type: "outline" /* OUTLINE */,
                });
                circles.push({
                    location: location_2,
                    type: "outer" /* OUTER */,
                });
                circles.push({
                    location: location_2,
                    type: "inner" /* INNER */,
                });
            }
            return circles;
        });
        this.getSizeScale = reselect_1.createSelector([getMaxLocationCircleSize, this.getMaxLocationMaxAbsTotal], function (maxLocationCircleSize, maxTotal) {
            var scale = d3_scale_1.scalePow()
                .exponent(1 / 2)
                .domain([0, maxTotal])
                .range([0, maxTotal > 0 ? maxLocationCircleSize : 1]);
            return function (v) { return scale(Math.abs(v)) || 0; };
        });
        this.getLocationCircleRadiusGetter = reselect_1.createSelector([
            this.getSizeScale,
            this.getLocationTotalInGetter,
            this.getLocationTotalOutGetter,
            this.getLocationTotalWithinGetter,
        ], function (sizeScale, getLocationTotalIn, getLocationTotalOut, getLocationTotalWithin) {
            return function (_a) {
                var location = _a.location, type = _a.type;
                var getSide = type === "inner" /* INNER */ ? Math.min : Math.max;
                var totalIn = getLocationTotalIn(location);
                var totalOut = getLocationTotalOut(location);
                var totalWithin = getLocationTotalWithin(location);
                var r = sizeScale(getSide(Math.abs(totalIn + totalWithin), Math.abs(totalOut + totalWithin)));
                if (type === "outline" /* OUTLINE */) {
                    return r + CIRCLE_OUTLINE_THICKNESS;
                }
                return r;
            };
        });
        this.getLocationCircleColorGetter = reselect_1.createSelector([
            this.getColors,
            getHighlightedLocationId,
            this.getLocationTotalInGetter,
            this.getLocationTotalOutGetter,
            this.getLocationTotalWithinGetter,
        ], function (colors, highlightedLocationId, getLocationTotalIn, getLocationTotalOut, getLocationTotalWithin) {
            var getLocationId = _this.inputAccessors.getLocationId;
            return function (_a) {
                var location = _a.location, type = _a.type;
                var isHighlighted = highlightedLocationId && highlightedLocationId === getLocationId(location);
                var isDimmed = highlightedLocationId && highlightedLocationId !== getLocationId(location);
                var totalWithin = getLocationTotalWithin(location);
                var totalIn = getLocationTotalIn(location) + totalWithin;
                var totalOut = getLocationTotalOut(location) + totalWithin;
                var isIncoming = type === "outer" /* OUTER */ && Math.abs(totalIn) > Math.abs(totalOut);
                var isPositive = (isIncoming === true && totalIn >= 0) || totalOut >= 0;
                var circleColors = (colors_1.isDiffColorsRGBA(colors) ? (isPositive ? colors.positive : colors.negative) : colors)
                    .locationCircles;
                if (isHighlighted && type === "outline" /* OUTLINE */) {
                    return circleColors.highlighted;
                }
                if (isDimmed) {
                    if (type === "outline" /* OUTLINE */) {
                        return colors_1.getDimmedCircleOutlineColor(colors.outlineColor, colors.dimmedOpacity);
                    }
                    return colors_1.getDimmedCircleColor(circleColors.inner, colors.dimmedOpacity);
                }
                if (type === "outline" /* OUTLINE */) {
                    return isIncoming ? circleColors.incoming : circleColors.inner;
                }
                if (type === "inner" /* INNER */) {
                    return circleColors.inner;
                }
                if (isIncoming === true) {
                    return circleColors.incoming;
                }
                return circleColors.outgoing;
            };
        });
        this.isLocationConnectedGetter = reselect_1.createSelector([this.getFilteredFlows, getHighlightedLocationId, getHighlightedFlow, getSelectedLocationIds], function (flows, highlightedLocationId, highlightedFlow, selectedLocationIds) {
            var _a = _this.inputAccessors, getFlowOriginId = _a.getFlowOriginId, getFlowDestId = _a.getFlowDestId;
            if (highlightedLocationId) {
                var isRelated = function (flow) {
                    var originId = getFlowOriginId(flow);
                    var destId = getFlowDestId(flow);
                    return (originId === highlightedLocationId ||
                        (selectedLocationIds && selectedLocationIds.indexOf(originId) >= 0) ||
                        destId === highlightedLocationId ||
                        (selectedLocationIds && selectedLocationIds.indexOf(destId) >= 0));
                };
                var locations_1 = new Set();
                for (var _i = 0, flows_1 = flows; _i < flows_1.length; _i++) {
                    var flow = flows_1[_i];
                    if (isRelated(flow)) {
                        locations_1.add(getFlowOriginId(flow));
                        locations_1.add(getFlowDestId(flow));
                    }
                }
                return function (id) { return locations_1.has(id); };
            }
            return function () { return false; };
        });
        this.getLocationAreaFillColorGetter = reselect_1.createSelector([this.getColors, getSelectedLocationIds, this.isLocationConnectedGetter], function (colors, selectedLocationIds, isLocationConnected) {
            return function (location) {
                var locationId = _this.inputAccessors.getLocationId(location);
                if (selectedLocationIds && selectedLocationIds.indexOf(locationId) >= 0) {
                    return colors.locationAreas.selected;
                }
                if (isLocationConnected(locationId)) {
                    return colors.locationAreas.connected;
                }
                return colors.locationAreas.normal;
            };
        });
    }
    Selectors.prototype.getFlowLinesColorGetter = function (colors, flowColorScale, highlighted, dimmed) {
        var _a = this.inputAccessors, getFlowMagnitude = _a.getFlowMagnitude, getFlowColor = _a.getFlowColor;
        return function (flow) {
            if (getFlowColor) {
                var color = getFlowColor(flow);
                if (color) {
                    return colors_1.colorAsRgba(color);
                }
            }
            if (highlighted) {
                if (colors_1.isDiffColorsRGBA(colors)) {
                    var positiveColor = colors.positive.flows.highlighted;
                    var negativeColor = colors.negative.flows.highlighted;
                    var magnitude = getFlowMagnitude(flow);
                    return magnitude >= 0 ? positiveColor : negativeColor;
                }
                else {
                    return colors.flows.highlighted;
                }
            }
            else {
                var magnitude = getFlowMagnitude(flow);
                var color = flowColorScale(magnitude);
                if (dimmed) {
                    return colors_1.getDimmedColor(color, colors.dimmedOpacity);
                }
                return color;
            }
        };
    };
    Selectors.prototype.setInputAccessors = function (inputAccessors) {
        this.inputAccessors = inputAccessors;
    };
    Selectors.prototype.getInputAccessors = function () {
        return this.inputAccessors;
    };
    return Selectors;
}());
exports.default = Selectors;
//# sourceMappingURL=Selectors.js.map